import os
import ast
import sys
import subprocess
from traceback import format_exc
sys.path.insert(0, os.path.abspath(os.pardir))
from pydetector.astexport import export_dict  # noqa: E402

__all__ = ['check_ast']

# This will store the AST to use for the "other" Python. It will be
# loaded the first time its tested by getOtherPythonAstEngine
OTHERPYTHON_AST = None

if sys.version_info[0] not in (2, 3):
    raise Exception('Sorry, only python 2 and 3 is supported')
PYMAJOR_CURRENT = sys.version_info[0]
PYMAJOR_OTHER = 2 if PYMAJOR_CURRENT == 3 else 3


def check_ast(code, try_other_on_sucess=False, verbosity=0,
              py2_exec='python2', py3_exec='python3'):
    """
    Try with the ast.parse of both Python 2 and 3 and then
    iterate over the retrieved AST to find specific syntax elements.

    Args:
        code (str) the source code to extract the AST for.

        try_other_on_sucess (bool): if the AST parser of the Python version that is
            running this script work try anyway with the other interpreter (True) or
            not (False). Please note that enabling this will usually increase a lot
            the "detection" rate of files as the version of the current interpreter.

        verbosity (int): level from 0 to 2. > 1 will show exceptions when parsing
            the AST.

        py2_exec (str): path or name (if in PATH) of the Python 2 interpreter when running
            this under Python 3.

        py3_exec (str): path or name (if in PATH) of the Python 3 interpreter to use when
            running this under Python 2.
    """

    current_ok = other_ok = False
    current_ast = other_ast = None
    py2_error = py3_error = current_error = other_error = ""

    pyexec_other = py2_exec if PYMAJOR_OTHER == 2 else py3_exec

    try:
        current_ast = export_dict(code)
        current_ok = True
    except:
        # current_ok remains false
        current_error = format_exc()
        if verbosity > 1:
            print('>>>> ASTCHECK: exception while parsing AST with Python%d:\n%s\n<<<< exception output end'
                  % (PYMAJOR_CURRENT, current_error))

    if verbosity:
        print('AST extractable with version %d?: %s' % (PYMAJOR_CURRENT, str(current_ok)))

    if not current_ok or try_other_on_sucess:
        # Open an external interpreter and try to export its AST
        cmd = [pyexec_other, "-c",
               "import ast,pydetector.astexport,sys;"
               "r=sys.stdin.read();"
               "print(pydetector.astexport.export_dict(r))"]

        if verbosity > 1:
            print('Running in other Python:\n%s' % ' '.join(cmd))

        try:
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            out, err = p.communicate(code.encode('utf-8'))
            if p.returncode == 0:
                if PYMAJOR_CURRENT == 3:
                    out = out.decode('utf-8')
                other_ast = ast.literal_eval(out)
                other_ok = True
            else:
                other_error = err
                if verbosity > 1:
                    print('>>>> ASTCHECK: error while parsing AST with Python%d:\n%s\n<<<< error output end'
                          % (PYMAJOR_OTHER, err))
        except:
            other_ok = False
            other_error = format_exc()
            if verbosity > 1:
                print('>>>> ASTCHECK: exception while parsing AST with Python%d:\n%s\n<<<< exception output end'
                      % (PYMAJOR_OTHER, other_error))

    if verbosity:
        print('AST extractable with version %d?: %s' % (PYMAJOR_OTHER, str(other_ok)))

    if PYMAJOR_CURRENT == 2:
        py2_ast   = current_ast
        py2_error = current_error
        py3_ast   = other_ast
        py3_error = other_error
    else:
        py3_ast   = current_ast
        py3_error = current_error
        py2_ast   = other_ast
        py2_error = other_error

    version = 0
    if current_ok and not other_ok:
        version = PYMAJOR_CURRENT
    elif other_ok and not current_ok:
        version = PYMAJOR_OTHER
    elif current_ok and other_ok:
        version = 6

    return version, py2_ast, py3_ast, py2_error, py3_error



import re

from io import open
from pydetector.ast_checks import check_ast
from pydetector.regexp_checks import check_syntax_regex, check_modules_regex,\
        check_modulesymbols_regex

__all__ = ['detect']

QUOTE_TRIPLE_SUBREGEX = re.compile(r'''\"{3}(.*?)\"{3}|'{3}(.*?)'{3}''', re.DOTALL)
QUOTE_SUBREGEX = re.compile(
    r'''(?<!(\\|'))\".*(?<!\\)\"''' +
    '|' +
    r"""(?<!(\\|\"))'.*(?<!\\)'"""
)
COMMENT_SUBREGEX = re.compile(r"#.*$", re.MULTILINE)


def remove_str_comments(code):
    """
    Remove all the comments in the code (from the # to EOL). Returns
    a new string with the comments removed. This improves the detection
    rates a lot removing most of the false positives.
    """
    # Empty strings
    newcode = QUOTE_TRIPLE_SUBREGEX.sub("''", code)
    newcode = QUOTE_SUBREGEX.sub("''", newcode)
    # Remove comments
    return COMMENT_SUBREGEX.sub("", newcode)


def detect(files=None, codestr=None, ast_checks=True, modules_checks=True,
        modsyms_checks=False, stop_on_ok_ast=False, modules_score=150,
        symbols_score=100, verbosity=0):
    """
        Try to detect if a source file is Python 2 or 3. It uses a combination of
    tests based on AST extraction and regular expressions.

    Args: files (List[str], optional): list of files. You can omit this parameter
    if you pass codestr.

        codestr (str, optional): source code of a single module to parse. You can
        omit this parameter if you pass "files".

        ast_checks (bool): enable checking if the AST parses with both Python
        versions

        modules_checks (bool): enable checking version-specific module imports

        modsyms_checks (bool): enable checking version-specific module symbols.
        Please note that this test can be much slower than the others.

        stop_on_ok_ast (bool): if the first AST tested works, don't even try with
        the other version

        modules_score (int): score given to specific-module matches

        symbols_score (int): score given to symbol-specific matches

        verbosity (int): verbosity level from 0 (quiet) to 2

    Return:
        Dictionary where each key is the filename and the value another dictionary
        with the keys "py2ast" and "py3ast" that will hold the AST if sucessfully
        parser for that version or "None", "version" with the version number (2 or
        3) or 6 is the module seems to be compatible with both versions, "matches"
        that will hold a list of the matched rules and scores and
        "py2_score/py3_score" with the specific score.
    """

    returndict = {}

    if not files:
        if not codestr:
            raise Exception('files or codestr parameters are required')
        files = ['<code_string>']

    for filename in files:
        returndict[filename] = {
            'py2ast': None,
            'py3ast': None,
            'version': 0,
            'matches': [],
            'py2_score': 0,
            'py3_score': 0,
            'py2_ast_errors': [],
            'py3_ast_errors': [],
        }
        retdict = returndict[filename]  # alias

        # helper for lazy bastards
        def apply_score(py2_score, py3_score):
            retdict['py2_score'] += py2_score
            retdict['py3_score'] += py3_score

        if verbosity:
            print('Checking file %s: ' % filename)

        input_code = None

        if filename == '<code_string>':
            input_code = codestr
        else:
            # this have problems if the file is not encoding in utf8 input_code = sys.stdin.read()
            # From most to  less common, this should cover 99.9% of the encodings used
            for encoding in ('utf_8', 'iso8859_15', 'iso8859_15', 'gb2313',
                    'cp1251', 'cp1252', 'cp1250', 'shift-jis', 'gbk', 'cp1256',
                    'iso8859-2', 'euc_jp', 'big5', 'cp874', 'euc_kr', 'iso8859_7'
                    'cp1255'):
                with open(filename, encoding=encoding) as infile:
                    try:
                        input_code = infile.read()
                        break
                    except UnicodeDecodeError:
                        continue
            else:
                raise Exception('Could not determine file encoding')

        if ast_checks:
            # Test the AST. This doesnt give points: either both pass, both fails
            # or one is correct and the other dont in which case we shortcircuit the return
            astversion, py2astroot, py3astroot, py2_err, py3_err = check_ast(
                        input_code, try_other_on_sucess=not stop_on_ok_ast,
                        verbosity=verbosity
            )
            retdict.update({
                # 'py2ast': {'PY2AST': py2astroot if py2astroot is not None else None},
                # 'py3ast': {'PY3AST': py3astroot if py2astroot is not None else None},
                'py2ast': {'PY2AST': py2astroot} if py2astroot else None,
                'py3ast': {'PY3AST': py3astroot} if py3astroot else None,
                'matches': [('PY%dASTOK' % astversion, ())]
            })
            if py2_err:
                retdict['py2_ast_errors'].append(py2_err)
            if py3_err:
                retdict['py3_ast_errors'].append(py3_err)

            # One parsed and the other didnt, no need to continue checking
            if astversion in (2, 3):
                retdict['version'] = astversion
                continue

        # Remove comments and emptyfy strings before doing the regex tests,
        # this will remove most fase positives
        cleaned_code = remove_str_comments(input_code)

        if modules_checks:
            apply_score(*check_syntax_regex(cleaned_code, retdict['matches']))
            apply_score(*check_modules_regex(cleaned_code, retdict['matches'],
                match_score = modules_score))

        # This one is SLOOOOOW
        if modsyms_checks:
            apply_score(
                *check_modulesymbols_regex(cleaned_code, retdict['matches'], symbols_score)
            )

        if retdict['py2_score'] > retdict['py3_score']:
            retdict['version'] = 2
        elif retdict['py3_score'] > retdict['py2_score']:
            retdict['version'] = 3
        else:
            retdict['version'] = 6

        if verbosity:
            print('Python 2 score: %d' % retdict['py2_score'])
            print('Python 3 score: %d' % retdict['py3_score'])
            print('\n')

    return returndict


if __name__ == '__main__':
    from pprint import pprint
    import sys
    with open('astexport.py') as f:
        code = f.read()
    pprint(detect(codestr=code, stop_on_ok_ast=False))



import sys
import argparse
import subprocess
from pprint import pprint
from pydetector.detector import detect

def parse_args():
    # TODO: add arguments for python executables
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbosity", type=int, default=0,
            help="increase output verbosity (0 to 2)")

    parser.add_argument("-d", "--defaultversion", type=int, default=0,
            help="Python version to return if there is a match. If not used "
                 "it will be just reported as a match (default=report matches)")

    parser.add_argument("-a", "--testast", action="store_true", default=True,
            help="Do the AST extraction test (default=enabled)")

    parser.add_argument("-o", "--asttestboth", dest='asttestboth', action='store_true',
            help="Do the AST test with the other version even if the first one works"
                 "(default=enabled)")
    parser.add_argument("-n", "--no-asttestboth", dest='asttestboth', action='store_false',
            help="Do the AST test with the other version even if the first one works"
                 "(default=enabled)")

    parser.add_argument("-m", "--testmodules", action="store_true", default=True,
            help="Test for version-specific modules (default=enabled)")

    parser.add_argument("-s", "--testmodulesyms", action="store_true", default=False,
            help="Test for version-specific module symbols (WARNING: SLOW!) (default=disabled)")

    parser.add_argument("-A", "--showast", action="store_true", default=False,
            help="Include the parsed AST")

    parser.add_argument("files", nargs=argparse.REMAINDER, help="Files to parse")

    args = parser.parse_args()

    if not args.files:
        print('List of files to operate on missing')
        parser.print_help()
        exit(1)

    if args.testast and args.verbosity > 0:
        PYMAJOR_CURRENT = sys.version_info[0]
        PYMAJOR_OTHER = 2 if PYMAJOR_CURRENT == 3 else 3
        if args.verbosity:
            print('Running under Python%d, Python%d will be used for the '
                   % (PYMAJOR_CURRENT, PYMAJOR_OTHER) + 'alternative AST tests.')

        # Test that the other Python version have pydetect installed
        try:
            # TODO: change the python exec when I implement the option
            # to specify the interpreter paths
            subprocess.check_call(['python%d' % PYMAJOR_OTHER, '-c',
                'from pydetector import ast_checks'])
        except subprocess.CalledProcess:
            print('Error: AST checks enabled but pydetector is not installed for ' +
                    'Python%d.\nPlease install it and try again or disable AST checks'
                    % PYMAJOR_OTHER)

    return args

def main():
    args = parse_args()

    returndict = detect(
            args.files,
            ast_checks=args.testast,
            modules_checks=args.testmodules,
            modsyms_checks=args.testmodulesyms,
            stop_on_ok_ast=not args.asttestboth,
            verbosity=args.verbosity
            )

    if not args.showast:
        for fdata in returndict:
            del returndict[fdata]['py2ast'] # not json serializable in the current form
            del returndict[fdata]['py3ast'] # not json serializable in the current form

    pprint(returndict)

    if args.verbosity:
        py2_count = py3_count = pyany_count = 0
        for key in returndict:
            version = returndict[key]['version']

            if version == 2 or (version == 6 and args.defaultversion == 2):
                py2_count += 1
            elif version == 3 or (version == 6 and args.defaultversion == 3):
                py3_count += 1
            else:
                pyany_count += 1

        print('%d files parsed, py2: %d, py3: %d any: %d' %
                (len(returndict), py2_count, py3_count, pyany_count))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

description = long_descr = "Python command line application & package to guess the Python version of a file"

main_ns = {}
with open("pydetector/version.py") as ver_file:
    exec(ver_file.read(), main_ns)

version = main_ns['__version__']
setup(
    name = "pydetector",
    version = version,

    description = description,
    long_description = long_descr,
    license = "Apache License 2.0",
    url = "https://github.com/juanjux/python-pydetector",
    download_url = "https://github.com/juanjux/python-pydetector/archive/%s.tar.gz" % version,
    author = "Juanjo Alvarez",
    author_email = "juanjo@juanjoalvarez.net",
    packages = find_packages(exclude=["tests"]),
    entry_points = {
        "console_scripts": [
            "pydetector = pydetector.cli:main"
        ]
    },
    install_requires = ["six"],
    extras_require = {},
    classifiers = [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Compilers",
        "Topic :: Software Development :: Disassemblers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5"
    ]
)



"""
Export an improved version of the Python AST for the given codestr
as a Python Dictionary or JSON object
"""
from __future__ import print_function

import ast
import tokenize
import token as token_module
from codecs import encode
from six import StringIO
# from pprint import pprint

# TODO: add an option to not change the node names of NameConstant, Num and Str


def export_dict(codestr):
    """ Returns the AST as a Python dictionary """
    visitor = DictExportVisitor(codestr)
    return visitor.parse()


def export_json(codestr, pretty_print=False):
    """ Returns the AST as a JSON object """
    import json
    dict_ = {'AST': export_dict(codestr)}
    json_ = json.dumps(dict_, indent=2 if pretty_print else 0, ensure_ascii=False)
    return json_, dict_


NAMECOUNTER = 0


def export_graphviz(codestr):
    """
    WARNING: Experimental, only compatible with Python 3 and unfinished
    """
    # TODO: use fillcolor for different ast_types
    # TODO: use shape for giving different shapes to different nodes: http://www.graphviz.org/content/node-shapes#polygon
    parent_nodekeys = {'body', 'names', 'targets', 'value', 'func', 'args', 'keywords',
                       'keys', 'left', 'right'}
    from graphviz import Graph

    # These nodes will show their childs as hanging from their parent
    def generate(dot, obj, parent=None, reparent=False):
        global NAMECOUNTER

        NAMECOUNTER += 1
        nodename = str(NAMECOUNTER)

        def get_name(node):
            subname = ''
            if isinstance(node, dict):
                ast = node.get('ast_type', '')
                subname = node.get('name', '')
                id = node.get('id', '')
                arg = node.get('arg', '')
                separator = ': ' if ast and (subname or id or arg) else ''
                name = '%s%s%s%s%s' % (ast, separator, subname, id, arg)
            else:
                name = str(type(node))
            return name, subname

        name, origname = get_name(obj)
        if reparent or type(obj) in (list, tuple):
            nodename = parent
        else:
            dot.node(nodename, label=name)
            if parent:
                dot.edge(parent, nodename)

        if isinstance(obj, dict):
            for childkey in parent_nodekeys:
                if childkey in obj:
                    reparent = childkey == 'names'
                    generate(dot, obj[childkey], nodename, reparent)

        elif isinstance(obj, list) or isinstance(obj, tuple):
            for value in obj:
                generate(dot, value, nodename)

    dict_ = export_dict(codestr),
    dot = Graph(comment="Python AST", format='pdf',
                graph_attr={'ranksep': '2.5'
                            })
    generate(dot, dict_, 'RootNode')

    with open('ast.dot', 'w') as dotfile:
        print(dot.source, file=dotfile)


TOKEN_TYPE     = 0
TOKEN_VALUE    = 1
TOKEN_STARTLOC = 2
TOKEN_ENDLOC   = 3
TOKEN_RAWVALUE = 4

ROW = 0
COL = 1

NOOP_TOKENS_LINE = {'COMMENT', 'INDENT', 'NL', 'NEWLINE'}


def _token_name(token):
    return token_module.tok_name[token[TOKEN_TYPE]]


class NoopExtractor(object):
    """
    Tokenize the source code and extract lines with tokens that Python's
    AST generator ignore like blanks and comments
    """

    def __init__(self, codestr):
        tokens = tokenize.generate_tokens(StringIO(codestr).readline)

        self.current_line = None
        self.all_lines = self._create_tokenized_lines(tokens, codestr)
        self.astmissing_lines = self._create_astmissing_lines()

        # This set is used to avoid adding the "same line-remainder noops" nodes as a child
        # of every "real" node to avoid having this node duplicated on all semantic
        # nodes in the same line, thus avoiding duplication. It will contain just the
        # line numbers of already added sameline_noops
        self._sameline_added_noops = set()

    def _create_astmissing_lines(self):
        """
        Return a copy of line_tokens containing lines ignored by the AST
        (comments and blanks-only lines)
        """
        lines = []
        nl_token = (token_module.NEWLINE, '\n', (0, 0), (0, 0), '\n')

        tname = _token_name
        for i, linetokens in enumerate(self.all_lines):
            if len(linetokens) == 1 and tname(linetokens[0]) == 'NL':
                lines.append(nl_token)
            else:
                for token in linetokens:
                    if tname(token) == 'COMMENT' and \
                            token[TOKEN_RAWVALUE].lstrip().startswith('#'):
                        lines.append(token)
                        break
                else:
                    lines.append(None)
        assert len(lines) == len(self.all_lines)

        for i, linetokens in enumerate(lines):
            if linetokens:
                self.current_line = i
                break
        else:
            self.current_line = len(lines)
        return lines

    def _create_tokenized_lines(self, tokens, codestr):
        """
        Create a list of tokenized lines
        """
        lines = codestr.splitlines() if codestr else []
        result = []
        for i in range(0, len(lines) + 1):
            result.append([])

        tname = _token_name
        for token in tokens:
            # Save noops in the line of the starting row except for strings where
            # we save it in the last line (because they can be multiline)
            if tname(token) == 'STRING':
                line = token[TOKEN_ENDLOC][ROW] - 1
            else:
                line = token[TOKEN_STARTLOC][ROW] - 1
            result[line].append(token)
        assert len(lines) + 1 == len(result), len(result)
        return result

    def previous_nooplines(self, node):
        """Return a list of the preceding comment and blank lines"""
        previous = []
        noop_first_lineno = None
        noop_last_lineno = None

        if hasattr(node, 'lineno'):
            while self.current_line < node.lineno:
                token = self.astmissing_lines[self.current_line]
                if token:
                    s = token[TOKEN_RAWVALUE].rstrip() + '\n'
                    previous.append(s)

                    # take only the first line of the noops as the start and the last
                    # one (overwriteen every iteration)
                    if not noop_first_lineno:
                        noop_first_lineno = self.current_line + 1
                    noop_last_lineno = self.current_line + 1
                self.current_line += 1
        return previous, noop_first_lineno, noop_last_lineno

    def remainder_noops_sameline(self, node):
        """
        Return a string containing the trailing (until EOL) noops for the
        node, if any. The ending newline is implicit and thus not returned
        """

        # Without a line number for the node we can't know
        if not hasattr(node, 'lineno'):
            return ''

        # Skip remainder comments already added to a node in this line to avoid every node
        # in the same line having it (which is not conceptually wrong, but not DRY)
        if node.lineno in self._sameline_added_noops:
            return ''

        # Module nodes have the remaining comments but since we put their first line as "1"
        # any comment on the first line would wrongly show as sameline comment for the module
        if node.__class__.__name__ == 'Module':
            return ''

        tokens = self.all_lines[node.lineno - 1]
        trailing = []

        tname = _token_name
        for token in tokens:
            if tname(token) not in NOOP_TOKENS_LINE:
                # restart
                trailing = []
            else:
                trailing.append({
                    'rowstart' : token[TOKEN_STARTLOC][ROW],
                    'colstart' : token[TOKEN_STARTLOC][COL],
                    'rowend'   : token[TOKEN_ENDLOC][ROW],
                    'colend'   : token[TOKEN_ENDLOC][COL],
                    'value'    : token[TOKEN_VALUE],
                    })
        if not trailing:
            return ''

        self._sameline_added_noops.add(node.lineno)
        nonewline_trailing = trailing[:-1] if trailing[-1]['value'] == '\n' else trailing
        return nonewline_trailing

    def remainder_noops(self):
        """return any remaining ignored lines."""
        trailing = []
        noop_last_lineno = None
        i = self.current_line
        noop_first_lineno = self.current_line + 1

        while i < len(self.astmissing_lines):
            token = self.astmissing_lines[i]
            if token:
                s = token[TOKEN_RAWVALUE]
                trailing.append(s)

            i += 1
            noop_last_lineno = i
        self.current_line = i
        return trailing, noop_first_lineno, noop_last_lineno


def node_dict(node, newdict, ast_type=None):
    """
    Shortcut that adds ast_type (if not specified),
    lineno and col_offset to the node-derived dictionary
    """
    if ast_type is None:
        ast_type = node.__class__.__name__

    newdict["ast_type"] = ast_type
    if hasattr(node, "lineno"):
        newdict["lineno"] = node.lineno
    if hasattr(node, "col_offset"):
        newdict["col_offset"] = node.col_offset

    return newdict


class DictExportVisitor(object):
    ast_type_field = "ast_type"

    def __init__(self, codestr, ast_parser=ast.parse, tsync_class=NoopExtractor):
        """
        Initialize the Token Syncer composited object, parse the source code
        and start visiting the node tree to add comments and other modifications

        Args:
            codestr (string): the string with the source code to parse and visit.

            ast_parser (function, optional): the AST parser function to use. It needs to take
            a string with the code as parameter. By default it will be ast.parse from stdlib.

            tsync_class (class, optional): the class to use to sinchronize the tokenizer with
            the AST visits. This is needed to extract aditional info like comments or whitespace
            that most AST parsers doesn't include in the tree. This class need to provide the public
            methods "previous_nooplines",  "remainder_noops_sameline" and "rmainder_noops".
            By default astexport.NoopExtractor will be used.
        """
        self.codestr = codestr
        self.sync = tsync_class(codestr)
        # this is used to store the parent node of the current one; currently is only
        # used to inherit the lineno and col_offset of the parent for "arguments" and
        # "Operator" types
        self._current_parent = None

    def _addNoopFields(self, node, visit_dict, root):
        if not isinstance(visit_dict, dict):
            return visit_dict

        def _create_nooplines_list(startline, noops_previous):
            nooplines = []
            curline = startline
            for noopline in noops_previous:
                nooplines.append({
                    'ast_type': 'NoopLine',
                    'noop_line': noopline,
                    'lineno': curline,
                    'col_offset': 1,
                })
                curline += 1
            return nooplines

        # Add all the noop (whitespace and comments) lines between the
        # last node and this one
        noops_previous, startline, endline = self.sync.previous_nooplines(node)
        if noops_previous:
            visit_dict['noops_previous'] = {
                "ast_type": "PreviousNoops",
                "lineno": startline,
                "col_offset": 1,
                "end_lineno": endline,
                "lines": _create_nooplines_list(startline, noops_previous)
            }

        # Other noops at the end of its significative line except the implicit
        # finishing newline
        noops_sameline = self.sync.remainder_noops_sameline(node)
        joined_sameline = ''.join([x['value'] for x in noops_sameline])
        if noops_sameline:
            visit_dict['noops_sameline'] = {
                "ast_type": "SameLineNoops",
                "lineno": node.lineno if hasattr(node, "lineno") else 0,
                "col_offset": noops_sameline[0]["colstart"],
                "col_end": noops_sameline[-1]["colend"],
                "noop_line": joined_sameline
            }

        # Finally, if this is the root node, add all noops after the last op node
        if root:
            noops_remainder, startline, endline = self.sync.remainder_noops()
            if noops_remainder:
                visit_dict['noops_remainder'] = {
                    "ast_type": "RemainderNoops",
                    "lineno": startline,
                    "col_offset": 1,
                    "end_lineno": endline,
                    "lines": _create_nooplines_list(startline, noops_remainder)
                }

    def parse(self):
        node = ast.parse(self.codestr, mode='exec')
        res = self.visit(node, root=True)
        return res

    def visit(self, node, root=False):
        node_type = node.__class__.__name__

        if node_type == 'Module':
            # add line and col since Python doesnt adds them
            node.__dict__['lineno'] = 1
            node.__dict__['col_offset'] = 0

        # the ctx property always has a "Load"/"Store"/etc nodes that
        # can be perfectly converted to a string value since they don't
        # hold anything more than the name
        if hasattr(node, 'ctx'):
            node.ctx = node.ctx.__class__.__name__

        meth = getattr(self, "visit_" + node_type, self.visit_other)
        visit_result = meth(node)
        self._addNoopFields(node, visit_result, root)

        if 'col_offset' in visit_result:
            # Python AST gives a 0 based column, I prefer a 1-based one
            visit_result['col_offset'] += 1

        return visit_result

    def visit_other(self, node):
        node_type = node.__class__.__name__
        nodedict = node_dict(node, {}, ast_type = node_type)

        # Visit fields
        for field in node._fields:
            meth = getattr(self, "visit_" + node_type, self.visit_other_field)
            nodedict[field] = meth(getattr(node, field))

            # these must inherit their position from the parent
            # FIXME: move to a method
            if field in ('args', 'op', 'ops', 'alias', 'keywords'):
                if isinstance(nodedict[field], dict):
                    toprocess = [nodedict[field]]
                elif isinstance(nodedict[field], list):
                    toprocess = nodedict[field]
                else:
                    continue

                for proc in toprocess:
                    if 'lineno' not in proc and hasattr(node, 'lineno'):
                        proc['lineno'] = node.lineno
                    # doesnt make sense to inherit col_offset since it would be different

        # Visit attributes
        for attr in node._attributes:
            meth = getattr(self, "visit_" + node_type + "_" + attr, self.visit_other_field)
            nodedict[attr] = meth(getattr(node, attr))

        return nodedict

    def visit_other_field(self, node):
        if isinstance(node, ast.AST):
            return self.visit(node)
        elif isinstance(node, list) or isinstance(node, tuple):
            return [self.visit(x) for x in node]
        else:
            # string attribute
            return node

    def visit_str(self, node):
        """
        This visits str fields inside nodes (which are represented as keys
        in the node dictionary), not Str AST nodes
        """
        return str(node)

    def visit_Str(self, node):
        return node_dict(node, {"LiteralValue": node.s}, ast_type="StringLiteral")

    def visit_Bytes(self, node):
        try:
            s = node.s.decode()
            encoding = 'utf8'
        except UnicodeDecodeError:
            # try with base64
            s = encode(node.s, 'base64').decode().strip()
            encoding = 'base64'

        return node_dict(node, {"LiteralValue": s, "encoding": encoding}, ast_type="ByteLiteral")

    def visit_NoneType(self, node):
        return 'NoneLiteral'

    def visit_Global(self, node):
        # Python AST by default stores global and nonlocal variable names
        # in a "names" array of strings. That breaks the structure of everything
        # else in the AST (dictionaries, properties or list of objects) so we
        # convert those names to Name objects

        names_as_nodes = [{"ast_type": "Name",
                          "id": i,
                          "lineno": node.lineno,
                          "col_offset": node.col_offset} for i in node.names]
        return node_dict(node, {"names": names_as_nodes}, ast_type="Global")

    def visit_Nonlocal(self, node):
        # ditto
        names_as_nodes = [{"ast_type": "Name",
                          "id": i,
                          "lineno": node.lineno,
                          "col_offset": node.col_offset} for i in node.names]
        return node_dict(node, {"names": names_as_nodes}, ast_type="Nonlocal")

    def visit_NameConstant(self, node):
        if hasattr(node, 'value'):
            repr_val = repr(node.value)
            if repr_val in ('True', 'False'):
                return node_dict(node, {"LiteralValue": node.value}, ast_type="BoolLiteral")
            elif repr_val == 'None':
                return node_dict(node, {"LiteralValue": node.value}, ast_type="NoneLiteral")
        return str(node)

    def visit_Num(self, node):
        if isinstance(node.n, int):
            ret_dict = { "NumType": "int", "LiteralValue": node.n }
        elif isinstance(node.n, float):
            ret_dict = { "NumType": "float", "LiteralValue": node.n }
        elif isinstance(node.n, complex):
            ret_dict = {
                        "NumType": "complex",
                        "LiteralValue": {"real": node.n.real, "imaginary": node.n.imag},
                       }

        return node_dict(node, ret_dict, ast_type="NumLiteral")


if __name__ == '__main__':
    import sys

    with open(sys.argv[1]) as codefile:
        content = codefile.read()

    # content = "#firstcomment\n#secondcomment\nppass #trailing comment\n#middle\n#secondmiddle\npass\n#beforelast\n#lastcomment"
    # print(content)

    # from pprint import pprint
    # pprint(export_dict(content))
    # print(export_json(content, pretty_print=True)[0])
    export_graphviz(content)
